{*****************************************}
{    Schlurp                              }
{    (c) Pascal Peyremorte                }
{                                         }
{    p.peyremorte@dial.oleane.com         }
{*****************************************}
{
 15/06/99  V2.00
 16/06/99  V2.01
 08/07/99  V2.02
 04/08/99  V2.10}

unit UF_Shlurp;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, Menus, ExtCtrls, StdCtrls;

Const
  CaseMax = 6; // Damier de 7x7
  Bord    = 14;// Taille des bord (numros de case)
  {Valeurs reprsantant le contenu des cases du plateau}
  Vide = 0;    // Case vide
  J1   = 1;    // Case occupe par le joueur 1
  J2   = 2;    // Case occupe par le joueur 2

Type
  TDamier   = Array[0..CaseMax,0..CaseMax] of Byte;

  TCoup = Record
    XD, YD : Byte; {Case de dpart}
    XA, YA : Byte; {Case d'arrive}
  End;

  TEcrPrin = class(TForm)
    MainMenu1: TMainMenu;
    M_Jeu: TMenuItem;
    MJ_NouvPartie: TMenuItem;
    MJ__1: TMenuItem;
    MJ_Quitter: TMenuItem;
    Plateau: TDrawGrid;
    M_J1: TMenuItem;
    MJ1_Hum: TMenuItem;
    Infos: TPanel;
    Total1: TLabel;
    Total2: TLabel;
    M_J2: TMenuItem;
    MJ2_Hum: TMenuItem;
    APropos1: TMenuItem;
    Apropos2: TMenuItem;
    AQuiLeTour: TShape;
    Coups: TListBox;
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure PlateauClick(Sender: TObject);
    procedure PlateauDrawCell(Sender: TObject; Col, Row: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure MJ_QuitterClick(Sender: TObject);
    procedure MJ_NouvPartieClick(Sender: TObject);
    procedure MJ_JoueurClick(Sender: TObject);
    procedure PlateauMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormDestroy(Sender: TObject);
    procedure Apropos2Click(Sender: TObject);
    procedure CoupsDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
  private { Dclarations prives }

    {Gestion des DLL externes}
    ListeDll : TStringList;
    Procedure AjouteJoueurs;
    Procedure AffecteProc(Joueur:Integer; Ref:String);

    Procedure TraceCase(Col,Lig:byte);
    Procedure MetPion(X,Y, Coul:Byte);
    Procedure JouePion(C:TCoup);
    Procedure AjouteCoup(C:TCoup; RefCoul:Byte);
    Function PeutJouer(Joueur:Byte):Boolean;
  end;

var
  EcrPrin: TEcrPrin;

{--------------------------------------------------------------------------}

implementation

uses UF_Apropos, Uf_Result;

{$R *.DFM}

Type
  TFoncNom    = Function : ShortString; StdCall;
  TFoncTaille = Function : Integer; StdCall;
  TFoncCalc   = Function (Const Damier:TDamier; Joueur:Byte):TCoup; StdCall;

  TJoueur = Record
    Nom       : String;
    Dll       : THandle;
    ProcCalc  : TFoncCalc;
    Duree     : Integer;
    NbreCoups : Integer;
  End;

Const
  Info : Array[J1..J2] of TJoueur = (
   (Nom:'Joueur 1'; Dll:0; ProcCalc:Nil; Duree:0),
   (Nom:'Joueur 2'; Dll:0; ProcCalc:Nil; Duree:0));

Var
  Couleur : Array[0..J2+1] of integer=(clWindow,clAqua,clGreen,clRed);

  TailleCase   : Integer;  // Taille en pixel d'une case
  BordPion     : Integer;  // Marge en bord du pion et bord case

  Damier       : TDamier;  // Le plateau de jeu
  Joueur       : Byte;     // Le joueur dont on attend le coup

  AttendJoueur : Boolean;  // Indique attend jeu manuel
  Select       : Boolean;  // Indique choix case dpart fait
  CoupAct      : TCoup;    // Coup manuel  effectuer

Function Sup(X,Y:ShortInt):ShortInt; Assembler;
Asm
  mov   ah,X
  mov   al,Y
  cmp   al,ah
  jg    @Fin
  mov   al,ah
@Fin:
End;

Function Inf(X,Y:ShortInt):ShortInt; Assembler;
Asm
  mov   ah,X
  mov   al,Y
  cmp   al,ah
  jl    @Fin
  mov   al,ah
@Fin:
End;

Function Stats(Joueur:Byte):String;
Begin
  With Info[Joueur] do
    If NbreCoups<>0
      then Result:=Nom+' : '+IntToStr(Duree div NbreCoups)+' ms par coup'
      else Result:=Nom+' : n''a pas eu  jouer !';
End;

//---------------------------------------------------------------------------
// Cration, destruction, changement de taille
//---------------------------------------------------------------------------

procedure TEcrPrin.FormCreate(Sender: TObject);
begin
  Plateau.Canvas.Pen.color:=clblack;
  Plateau.Canvas.Pen.Width:=1;
  ListeDll:=TStringList.Create;

  AjouteJoueurs;
end;

procedure TEcrPrin.FormDestroy(Sender: TObject);
begin
  ListeDll.Free;
  With Info[J1] do
    If Dll<>0 then FreeLibrary(Dll);
  With Info[J2] do
    If Dll<>0 then FreeLibrary(Dll);
end;

procedure TEcrPrin.FormResize(Sender: TObject);
Var
   Bordure:Integer;
   TaillePlateau:Integer;
Begin
  Bordure:=Plateau.Width-Plateau.ClientWidth; // Encadrement fentre + panel

  TailleCase:=(ClientWidth-Bord) div (CaseMax+3)-Plateau.GridLineWidth;
  TaillePlateau:=Bord + TailleCase*(CaseMax+1)+ Plateau.GridLineWidth*(CaseMax+2);

  ClientWidth:=  TaillePlateau + TailleCase*2 + Bordure;
  ClientHeight:= TaillePlateau + Bordure;

  Plateau.ClientWidth:=TaillePlateau;
  Plateau.DefaultRowHeight:=TailleCase;
  Plateau.DefaultColWidth:=TailleCase;
  Plateau.ColWidths[0]:=14;
  Plateau.RowHeights[0]:=14;

  BordPion:=TailleCase div 10;
  Total1.Width:=Infos.ClientWidth-24;
  Total2.Width:=Infos.ClientWidth-24;
  Coups.Width:=Infos.ClientWidth-24;
  Coups.Height:= Infos.Height-165;
  AQuiLeTour.Width:=Infos.ClientWidth-24;
  AQuiLeTour.Top:=Infos.Height-49;
end;

procedure TEcrPrin.MJ_QuitterClick(Sender: TObject);
begin
  Close;
end;


//---------------------------------------------------------------------------
// Gestion des joueurs en Dll
//---------------------------------------------------------------------------

Procedure TEcrPrin.AjouteJoueurs;
Var
  F    : TSearchRec;
  HDll : THandle;
  Nom  : TFoncNom;
  Dim  : TFoncTaille;
  Item : TMenuItem;
Begin
  {Ajoute aux menus Joueurs les dll prsentes}
  If FindFirst('*.dll',faAnyFile,F)=0 then
  Repeat
    HDLL:=LoadLibrary(PChar(F.Name));
    If HDLL<>0 then
    Try {Dll correctement charge}
      @Nom:=GetProcAddress(HDll, 'GetPlayerName');
      @Dim:=GetProcAddress(HDll, 'GetBoardSize');
      If (@Nom<>Nil) and (@Dim<>Nil) and (Dim=CaseMax+1) then
      Begin
        {Garde rfrence du nom de la dll}
        ListeDll.Add(Nom+'='+F.Name);

        {Ajoute un lment aux menus J1 et J2}
        Item:=TMenuItem.Create(Self);
        Item.Caption:=Nom;
        Item.GroupIndex:=1;
        Item.RadioItem:=True;
        Item.OnClick:=MJ_JoueurClick;
        M_J1.Add(Item);

        Item:=TMenuItem.Create(Self);
        Item.Caption:=Nom;
        Item.GroupIndex:=1;
        Item.RadioItem:=True;
        Item.OnClick:=MJ_JoueurClick;
        M_J2.Add(Item);
      End;
    Finally
      FreeLibrary(HDLL);
    End;
  Until FindNext(F)<>0;
  FindClose(F);
End;

Procedure TEcrPrin.AffecteProc(Joueur:Integer; Ref:String);
Begin
  With Info[Joueur] do
  Begin
    If Dll<>0 then {Une Dll est dj charge, la libre}
    Begin
      FreeLibrary(Dll);
      Dll:=0;
    End;

    Nom:=Ref;
    If Nom='Humain' then Exit; // Dll=0 suffit pour tre en mode humain

    Dll:=LoadLibrary(PChar(ListeDll.Values[Ref]));
    ProcCalc:=GetProcAddress(Dll, 'ComputeMove');
  End;
End;

procedure TEcrPrin.MJ_JoueurClick(Sender: TObject);
begin
  With (Sender as TMenuItem) do
  Begin
    Checked:=True;  // Indique item est slectionn
    AffecteProc(Parent.Tag, Caption); // Met en place Dll
  End;
end;


//---------------------------------------------------------------------------
// Dessin du plateau
//---------------------------------------------------------------------------

procedure TEcrPrin.PlateauDrawCell(Sender: TObject; Col, Row: Integer;
  Rect: TRect; State: TGridDrawState);
Var
  S:String;
  X,Y:Integer;
begin
  If (Col=0) or (Row=0) then
  With Plateau.Canvas do
  Begin
    If Col<>0 then
      S:=Char(Col+48) {Premire range}
    else If Row<>0 then
      S:=Char(Row+64); {Premire colonne}
    X:=(Rect.Right-Rect.Left-TextWidth(S)) div 2  + Rect.Left;
    Y:=(Rect.Bottom-Rect.Top-TextHeight(S)) div 2 + Rect.Top;
    Brush.Color:=Plateau.Color;
    FillRect(Rect);
    TextOut(X,Y,S);
  End
  else
    TraceCase(Col,Row);
end;

Procedure TEcrPrin.TraceCase(Col,Lig:byte);
var
  Rect:TRect;
Begin
  Rect:=Plateau.CellRect(Col,Lig);
  With Plateau.canvas do
  Begin
    If Select and (Col=CoupAct.XD+1) and (Lig=CoupAct.YD+1)
      then Brush.Color:=clBtnFace
      else Brush.color:=clWindow;
    Pen.style:=psSolid;
    {Pen.Mode:=pmBlack;}
    Pen.color:=clBlack;
    Pen.width:=1;
    Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);

    If (Col>0) and (Lig>0) then
    Begin
      Dec(Col); Dec(Lig);
      If Damier[Col,Lig]>0 then
      Begin
        Brush.color:=Couleur[Damier[Col,Lig]];
        Ellipse(Rect.Left+BordPion, Rect.Top+BordPion,
                Rect.Right-BordPion,Rect.Bottom-BordPion);
      End;
    End;
  End;
End;

PROCEDURE TEcrPrin.MetPion(X,Y, Coul:Byte);
BEGIN
  Damier[X,Y]:=Coul;
  TraceCase(X+1,Y+1);
END;


//---------------------------------------------------------------------------
// Dessin de la liste
//---------------------------------------------------------------------------

procedure TEcrPrin.CoupsDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
Var S:String;
begin
  S:=Coups.Items[Index];
  With Coups.Canvas do
  Begin
    Brush.Color:=Couleur[Byte(S[1])];
    FillRect(Rect);
    If S[1]=#1
      then Font.Color:=clBlack
      else Font.Color:=clWhite;
    Delete(S,1,1);
    TextOut(Rect.Left+2,Rect.Top+1,S);
  End;
end;

Procedure TEcrPrin.AjouteCoup(C:TCoup; RefCoul:Byte);
Begin
  Coups.ItemIndex :=
    Coups.Items.Add(Char(RefCoul)+
                    Char(C.YD+65)+Char(C.XD+49)+'->'+
                    Char(C.YA+65)+Char(C.XA+49));
End;
//---------------------------------------------------------------------------
// Joue un pion...
//---------------------------------------------------------------------------

Procedure TEcrPrin.JouePion(C:TCoup);
Var
   I,J,K:Byte;
   Rect:TRect;
Begin
  AjouteCoup(C, Joueur);
  {Ote le pion qui saute ou remet sa couleur originale s'il reste}
  If (Abs(C.XA-C.XD)>1) or (Abs(C.YA-C.YD)>1)
    then MetPion(C.XD,C.YD,Vide)
    else MetPion(C.XD,C.YD,Joueur);

  {Ajoute le pion jou}
  MetPion(C.XA,C.YA,Joueur);

  Plateau.canvas.Brush.style:=bsClear;
  Plateau.canvas.Pen.Color:=Couleur[Damier[C.XA,C.YA]];

  {Phagocyte les case voisines}
  For K:=TailleCase div 2 -1 downto BordPion do
  Begin
    For I:=Sup(0,C.XA-1) to Inf(CaseMax,C.XA+1) do
      For J:=Sup(0,C.YA-1) to Inf(CaseMax,C.YA+1) do
        If Damier[I,J]=J1+J2-Joueur then
          If K>BordPion then
          Begin
            Rect:=Plateau.CellRect(I+1,J+1);
            Plateau.canvas.Ellipse(Rect.Left+K, Rect.Top+K,
                                   Rect.Right-K,Rect.Bottom-K);
            Sleep(10);
          End
          else
            MetPion(I,J,Joueur);
  End;
End;


//-----------------------------------------------------------------------------
// Slection d'un coup par un joueur
//-----------------------------------------------------------------------------

procedure TEcrPrin.PlateauMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
Var Col,Row:integer;
begin
  Plateau.MouseToCell(X,Y, Col,Row);
  Dec(Col); Dec(Row);
  If (Col>=0) and (Row>=0) and
     ((Damier[Col,Row]=Joueur) or
      (Select and (Damier[Col,Row]=Vide) and
       (abs(Col-CoupAct.XD)<=2) and (abs(Row-CoupAct.YD)<=2)))
    then Plateau.cursor:=crHandPoint
    else Plateau.cursor:=crNo;
end;

procedure TEcrPrin.PlateauClick(Sender: TObject);
Var
  C,R:Integer;
begin
  If Not AttendJoueur then Exit;

  C:=Plateau.Col-1; R:=Plateau.Row-1;
  With CoupAct do
  If Damier[C,R]=Joueur then
    {Choisit un de ses propres pions}
    If Select and (C=XD) and (R=YD) then
    Begin {Pion dj slectionn ! Annule slection}
      Select:=False;
      TraceCase(C+1,R+1);
    End
    else
    Begin {Slectionne ce pion comme case de dpart}
      If Select then
      Begin
        Select:=False;
        TraceCase(XD+1,YD+1);
      End;
      XD:=C;
      YD:=R;
      Select:=True;
      TraceCase(C+1,R+1);
    End
  else
    If Select and (abs(C-XD)<=2) and (abs(R-YD)<=2) and (Damier[C,R]=Vide) then
    Begin {et voil la case d'arrive}
      XA:=C;
      YA:=R;
      Select:=False;
      AttendJoueur:=False;
    End;
end;


//-----------------------------------------------------------------------------
// Lancement d'une partie
//-----------------------------------------------------------------------------

Function TEcrPrin.PeutJouer(Joueur:Byte):Boolean;
Var X,Y, L,C:byte;
Begin
  For X:=0 to CaseMax do   // Pour chaque ligne
    For Y:=0 to CaseMax do // et chaque colonne
      If Damier[X,Y]=Joueur then  //Cherche un pion du joueur
        For C:=Sup(0,X-2) to Inf(CaseMax,X+2) do
          For L:=Sup(0,Y-2) to Inf(CaseMax,Y+2) do
            If Damier[C,L]=Vide then
            Begin
              Result:=True;
              Exit;
            End;
  Result:=False;
End;

procedure TEcrPrin.MJ_NouvPartieClick(Sender: TObject);
Var Fini : Boolean;
    I,J  : Byte;
    T    : Array[Vide..J2] of Byte;   {Total des pions de chaque joueur}
    DamierExt:TDamier;
begin
  Plateau.Onclick:=PlateauClick;
  {Initialise la partie}
  FillChar(Damier,Sizeof(Damier),Vide);
  MetPion(0,0, J1);
  MetPion(0,CaseMax, J2);
  MetPion(CaseMax,0, J2);
  MetPion(CaseMax,CaseMax, J1);
  Joueur:=J1;   {Premier joueur = Bleu}
  Plateau.Invalidate; {Redessine le plateau}
  Coups.Clear;

  Info[J1].Duree:=0;  {Met les compteurs  0}
  Info[J2].Duree:=0;
  Info[J1].NbreCoups:=0;
  Info[J2].NbreCoups:=0;

  Repeat
    AQuiLeTour.Brush.Color:=Couleur[Joueur];
    Application.ProcessMessages;    // Pour afficher avant calcul

    {Attend proposition du joueur}
    With Info[Joueur] do
    Begin
      DamierExt:=Damier;
      Dec(Duree, GetTickCount);
      If Dll=0 then
      Begin {Attend slection du coup par oprateur}
        Select:=False;
        AttendJoueur:=True;
        Repeat
          Application.ProcessMessages;
          If Application.Terminated then Exit;
          If Dll<>0 then
          Begin
            CoupAct:=ProcCalc(DamierExt, Joueur);
            Break;
          End;
        until Not AttendJoueur;
        {Rsulat est dans CoupAct}
      End
      else {Demande  un module de choisir le coup}
        CoupAct:=ProcCalc(DamierExt, Joueur);
      Inc(Duree, GetTickCount);
      Inc(NbreCoups);
    End;

    With CoupAct do
    If (Damier[XD,YD]=Joueur) and (Damier[XA,YA]=Vide) and
       (Abs(XD-XA)<=2) and (Abs(YD-YA)<=2) then
    Begin
      {Joue le coup}
      JouePion(CoupAct);

      {Comptage pour affichage nombre pions et dtection fin partie}
      FillChar(T, SizeOf(T), 0);
      For I:=0 to CaseMax do
        For J:=0 to CaseMax do
          Inc(T[Damier[I,J]]);
      Total1.Caption:=IntToStr(T[J1]); // Affiche rsultats
      Total2.Caption:=IntToStr(T[J2]);
      Fini:= (T[Vide]=0) or           // Plus de cases vides, partie finie
             (T[J1]=0) or (T[J2]=0);  // Un joueur n'a plus de pions !

      {Contrle si suivant peut jouer ou doit passer}
      Joueur:=J1+J2-Joueur;           // Change de camp
      If Not PeutJouer(Joueur) then
      Begin                           // Le joueur doit passer
        Joueur:=J1+J2-Joueur;
        If Not PeutJouer(Joueur) then
          Fini:=True;                 // Les 2 jouers passent : partie finie !
      End;
    End
    else
    Begin
      AjouteCoup(CoupAct,3);
      ShowMessage(Info[Joueur].Nom+' a tent de tricher !');
      T[Joueur]:=0;
      Fini:=true;
    End;
  until Fini;
  Plateau.OnClick:=Nil;

  With EcrResultat do
  Begin
    If T[J1]>T[J2]     {Dtermine le gagnant}
      then Joueur:=J1
      else Joueur:=J2;
    lblGagnant.Caption:=UpperCase(Info[Joueur].Nom);
    lblGagnant.Color:=Couleur[Joueur];
    lblJ1.Caption:=Stats(J1);
    lblJ2.Caption:=Stats(J2);
    ShowModal;
  End;
  Plateau.Onclick:=MJ_NouvPartieClick;
end;

procedure TEcrPrin.Apropos2Click(Sender: TObject);
begin
  EcrApropos.ShowModal;
end;

end.

